import numpy as np

from pymoo.indicators.igd import IGD
from pymoo.termination.delta import DeltaToleranceTermination
from pymoo.util.normalization import normalize



class DesignSpaceTermination(DeltaToleranceTermination):

    def __init__(self, tol=0.005, **kwargs):
        """
        Check the distance in the design-space and terminate based on tolerance.
        (only works if variables can be converted to float)
        """
        super().__init__(tol, **kwargs)

    def _delta(self, prev, current):
        try:
            return IGD(current.astype(float)).do(prev.astype(float))
        except:
            return np.inf

    def _data(self, algorithm):

        X = algorithm.opt.get("X")

        # do normalization if bounds are given
        problem = algorithm.problem
        if X.dtype != object and problem.has_bounds():
            X = normalize(X, xl=problem.xl, xu=problem.xu)

        return X



import numpy as np
from scipy.spatial.distance import pdist, squareform
from pymoo.util.misc import find_duplicates
from pymoo.util.function_loader import load_function


def get_crowding_function(label):

    if label == "cd":
        fun = FunctionalDiversity(calc_crowding_distance, filter_out_duplicates=False)
    elif (label == "pcd") or (label == "pruning-cd"):
        fun = FunctionalDiversity(load_function("calc_pcd"), filter_out_duplicates=True)
    elif label == "ce":
        fun = FunctionalDiversity(calc_crowding_entropy, filter_out_duplicates=True)
    elif label == "mnn":
        fun = FuncionalDiversityMNN(load_function("calc_mnn"), filter_out_duplicates=True)
    elif label == "2nn":
        fun = FuncionalDiversityMNN(load_function("calc_2nn"), filter_out_duplicates=True)
    elif hasattr(label, "__call__"):
        fun = FunctionalDiversity(label, filter_out_duplicates=True)
    elif isinstance(label, CrowdingDiversity):
        fun = label
    else:
        raise KeyError("Crowding function not defined")
    return fun


class CrowdingDiversity:

    def do(self, F, n_remove=0):
        # Converting types Python int to Cython int would fail in some cases converting to long instead
        n_remove = np.intc(n_remove)
        F = np.array(F, dtype=np.double)
        return self._do(F, n_remove=n_remove)

    def _do(self, F, n_remove=None):
        pass


class FunctionalDiversity(CrowdingDiversity):

    def __init__(self, function=None, filter_out_duplicates=True):
        self.function = function
        self.filter_out_duplicates = filter_out_duplicates
        super().__init__()

    def _do(self, F, **kwargs):

        n_points, n_obj = F.shape

        if n_points <= 2:
            return np.full(n_points, np.inf)

        else:

            if self.filter_out_duplicates:
                # filter out solutions which are duplicates - duplicates get a zero finally
                is_unique = np.where(np.logical_not(find_duplicates(F, epsilon=1e-32)))[0]
            else:
                # set every point to be unique without checking it
                is_unique = np.arange(n_points)

            # index the unique points of the array
            _F = F[is_unique]

            _d = self.function(_F, **kwargs)

            d = np.zeros(n_points)
            d[is_unique] = _d

        return d


class FuncionalDiversityMNN(FunctionalDiversity):

    def _do(self, F, **kwargs):

        n_points, n_obj = F.shape

        if n_points <= n_obj:
            return np.full(n_points, np.inf)

        else:
            return super()._do(F, **kwargs)


def calc_crowding_distance(F, **kwargs):
    n_points, n_obj = F.shape

    # sort each column and get index
    I = np.argsort(F, axis=0, kind='mergesort')

    # sort the objective space values for the whole matrix
    F = F[I, np.arange(n_obj)]

    # calculate the distance from each point to the last and next
    dist = np.row_stack([F, np.full(n_obj, np.inf)]) - np.row_stack([np.full(n_obj, -np.inf), F])

    # calculate the norm for each objective - set to NaN if all values are equal
    norm = np.max(F, axis=0) - np.min(F, axis=0)
    norm[norm == 0] = np.nan

    # prepare the distance to last and next vectors
    dist_to_last, dist_to_next = dist, np.copy(dist)
    dist_to_last, dist_to_next = dist_to_last[:-1] / norm, dist_to_next[1:] / norm

    # if we divide by zero because all values in one columns are equal replace by none
    dist_to_last[np.isnan(dist_to_last)] = 0.0
    dist_to_next[np.isnan(dist_to_next)] = 0.0

    # sum up the distance to next and last and norm by objectives - also reorder from sorted list
    J = np.argsort(I, axis=0)
    cd = np.sum(dist_to_last[J, np.arange(n_obj)] + dist_to_next[J, np.arange(n_obj)], axis=1) / n_obj

    return cd


def calc_crowding_entropy(F, **kwargs):
    """Wang, Y.-N., Wu, L.-H. & Yuan, X.-F., 2010. Multi-objective self-adaptive differential 
    evolution with elitist archive and crowding entropy-based diversity measure. 
    Soft Comput., 14(3), pp. 193-209.

    Parameters
    ----------
    F : 2d array like
        Objective functions.

    Returns
    -------
    ce : 1d array
        Crowding Entropies
    """
    n_points, n_obj = F.shape

    # sort each column and get index
    I = np.argsort(F, axis=0, kind='mergesort')

    # sort the objective space values for the whole matrix
    F = F[I, np.arange(n_obj)]

    # calculate the distance from each point to the last and next
    dist = np.row_stack([F, np.full(n_obj, np.inf)]) - np.row_stack([np.full(n_obj, -np.inf), F])

    # calculate the norm for each objective - set to NaN if all values are equal
    norm = np.max(F, axis=0) - np.min(F, axis=0)
    norm[norm == 0] = np.nan

    # prepare the distance to last and next vectors
    dl = dist.copy()[:-1]
    du = dist.copy()[1:]

    # Fix nan
    dl[np.isnan(dl)] = 0.0
    du[np.isnan(du)] = 0.0

    # Total distance
    cd = dl + du

    # Get relative positions
    pl = (dl[1:-1] / cd[1:-1])
    pu = (du[1:-1] / cd[1:-1])

    # Entropy
    entropy = np.row_stack([np.full(n_obj, np.inf),
                            -(pl * np.log2(pl) + pu * np.log2(pu)),
                            np.full(n_obj, np.inf)])

    # Crowding entropy
    J = np.argsort(I, axis=0)
    _cej = cd[J, np.arange(n_obj)] * entropy[J, np.arange(n_obj)] / norm
    _cej[np.isnan(_cej)] = 0.0
    ce = _cej.sum(axis=1)

    return ce


def calc_mnn_fast(F, **kwargs):
    return _calc_mnn_fast(F, F.shape[1], **kwargs)


def calc_2nn_fast(F, **kwargs):
    return _calc_mnn_fast(F, 2, **kwargs)


def _calc_mnn_fast(F, n_neighbors, **kwargs):

    # calculate the norm for each objective - set to NaN if all values are equal
    norm = np.max(F, axis=0) - np.min(F, axis=0)
    norm[norm == 0] = 1.0

    # F normalized
    F = (F - F.min(axis=0)) / norm

    # Distances pairwise (Inefficient)
    D = squareform(pdist(F, metric="sqeuclidean"))

    # M neighbors
    M = F.shape[1]
    _D = np.partition(D, range(1, M+1), axis=1)[:, 1:M+1]

    # Metric d
    d = np.prod(_D, axis=1)

    # Set top performers as np.inf
    _extremes = np.concatenate((np.argmin(F, axis=0), np.argmax(F, axis=0)))
    d[_extremes] = np.inf

    return d



import numpy as np
from scipy.spatial.distance import pdist, squareform

def calc_mnn(X, n_remove=0):
    return calc_mnn_base(X, n_remove=n_remove, twonn=False)

def calc_2nn(X, n_remove=0):
    return calc_mnn_base(X, n_remove=n_remove, twonn=True)

def calc_mnn_base(X, n_remove=0, twonn=False):

    N = X.shape[0]
    M = X.shape[1]
    
    if N <= M:
        return np.full(N, np.inf)

    if n_remove <= (N - M):
        if n_remove < 0:
            n_remove = 0
        else:
            pass
    else:
        n_remove = N - M
    
    if twonn:
        M = 2

    extremes_min = np.argmin(X, axis=0)
    extremes_max = np.argmax(X, axis=0)
    
    min_vals = np.min(X, axis=0)
    max_vals = np.max(X, axis=0)

    extremes = np.concatenate((extremes_min, extremes_max))
    
    X = (X - min_vals) / (max_vals - min_vals)
    
    H = np.arange(N)
    
    D = squareform(pdist(X, metric="sqeuclidean"))
    Dnn = np.partition(D, range(1, M+1), axis=1)[:, 1:M+1]
    d = np.product(Dnn, axis=1)
    d[extremes] = np.inf
    
    n_removed = 0

    #While n_remove not acheived
    while n_removed < (n_remove - 1):

        #Obtain element to drop
        _d = d[H]
        _k = np.argmin(_d)
        k = H[_k]
        H = H[H != k]

        #Update index
        n_removed = n_removed + 1
        if n_removed == n_remove:
            break

        else:

            D[:, k] = np.inf
            Dnn[H] = np.partition(D[H], range(1, M+1), axis=1)[:, 1:M+1]
            d[H] = np.product(Dnn[H], axis=1)
            d[extremes] = np.inf
    
    return d




# distutils: language = c++
# cython: language_level=2, boundscheck=False, wraparound=False, cdivision=True

import numpy as np

from pymoo.cython.utils cimport c_get_drop, c_get_argmin, c_get_argmax, c_normalize_array

from libcpp cimport bool
from libcpp.vector cimport vector
from libcpp.set cimport set as cpp_set


cdef extern from "math.h":
    double HUGE_VAL


# Python definition
def calc_pcd(double[:, :] X, int n_remove=0):

    cdef:
        int N, M, n
        cpp_set[int] extremes
        vector[int] extremes_min, extremes_max
        int[:, :] I

    N = X.shape[0]
    M = X.shape[1]

    if n_remove <= (N - M):
        if n_remove < 0:
            n_remove = 0
        else:
            pass
    else:
        n_remove = N - M

    extremes_min = c_get_argmin(X)
    extremes_max = c_get_argmax(X)

    extremes = cpp_set[int]()

    for n in extremes_min:
        extremes.insert(n)

    for n in extremes_max:
        extremes.insert(n)

    _I = np.argsort(X, axis=0, kind='mergesort').astype(np.intc)
    I = _I[:, :]

    X = c_normalize_array(X, extremes_max, extremes_min)

    return c_calc_pcd(X, I, n_remove, N, M, extremes)


# Returns crowding metrics with recursive elimination
cdef c_calc_pcd(double[:, :] X, int[:, :] I, int n_remove, int N, int M, cpp_set[int] extremes):

    cdef:
        int n, n_removed, k
        cpp_set[int] calc_items
        cpp_set[int] H
        double[:, :] D
        double[:] d

    # Define items to calculate distances
    calc_items = cpp_set[int]()
    for n in range(N):
        calc_items.insert(n)
    for n in extremes:
        calc_items.erase(n)

    # Define remaining items to evaluate
    H = cpp_set[int]()
    for n in range(N):
        H.insert(n)

    # Initialize
    n_removed = 0

    # Initialize neighbors and distances
    _D = np.full((N, M), HUGE_VAL, dtype=np.double)
    dd = np.full((N,), HUGE_VAL, dtype=np.double)

    D = _D[:, :]
    d = dd[:]

    # Fill in neighbors and distance matrix
    c_calc_pcd_iter(
            X,
            I,
            D,
            N, M,
            calc_items,
        )

    # Obtain distance metrics
    c_calc_d(d, D, calc_items, M)

    # While n_remove not acheived
    while n_removed < (n_remove - 1):

        # Obtain element to drop
        k = c_get_drop(d, H)
        H.erase(k)

        # Update index
        n_removed = n_removed + 1

        # Get items to be recalculated
        calc_items = c_get_calc_items(I, k, M, N)
        for n in extremes:
            calc_items.erase(n)

        # Fill in neighbors and distance matrix
        c_calc_pcd_iter(
                X,
                I,
                D,
                N, M,
                calc_items,
            )

        # Obtain distance metrics
        c_calc_d(d, D, calc_items, M)

    return dd


# Iterate
cdef c_calc_pcd_iter(
    double[:, :] X,
    int[:, :] I,
    double[:, :] D,
    int N, int M,
    cpp_set[int] calc_items,
    ):

    cdef:
        int i, m, n, l, u

    # Iterate over items to calculate
    for i in calc_items:

        # Iterate over elements in X
        for m in range(M):

            for n in range(N):

                if i == I[n, m]:

                    l = I[n - 1, m]
                    u = I[n + 1, m]

                    D[i, m] = (X[u, m] - X[l, m]) / M


# Calculate crowding metric
cdef c_calc_d(double[:] d, double[:, :] D, cpp_set[int] calc_items, int M):

    cdef:
        int i, m

    for i in calc_items:

        d[i] = 0
        for m in range(M):
            d[i] = d[i] + D[i, m]


# Returns indexes of items to be recalculated after removal
cdef cpp_set[int] c_get_calc_items(
    int[:, :] I,
    int k, int M, int N
    ):

    cdef:
        int n, m
        cpp_set[int] calc_items

    calc_items = cpp_set[int]()

    # Iterate over all elements in I
    for m in range(M):

        for n in range(N):

            if I[n, m] == k:

                # Add to set of items to be recalculated
                calc_items.insert(I[n - 1, m])
                calc_items.insert(I[n + 1, m])

                # Remove element from sorted array
                I[n:-1, m] = I[n + 1:, m]

    return calc_items



import numpy as np

from pymoo.core.mutation import Mutation


class ChoiceRandomMutation(Mutation):

    def _do(self, problem, X, **kwargs):
        assert problem.vars is not None

        # ensure the type object (fixed string length <UX can cause issues)
        X = X.astype(object)

        prob_var = self.get_prob_var(problem, size=len(X))

        for k, (_, var) in enumerate(problem.vars.items()):
            mut = np.where(np.random.random(len(X)) < prob_var)[0]

            v = var.sample(len(mut))
            X[mut, k] = v

        return X




import numpy as np

from pymoo.core.sampling import Sampling


def random(problem, n_samples=1):
    X = np.random.random((n_samples, problem.n_var))

    if problem.has_bounds():
        xl, xu = problem.bounds()
        assert np.all(xu >= xl)
        X = xl + (xu - xl) * X

    return X


class FloatRandomSampling(Sampling):

    def _do(self, problem, n_samples, **kwargs):
        X = np.random.random((n_samples, problem.n_var))

        if problem.has_bounds():
            xl, xu = problem.bounds()
            assert np.all(xu >= xl)
            X = xl + (xu - xl) * X

        return X


class BinaryRandomSampling(Sampling):

    def _do(self, problem, n_samples, **kwargs):
        val = np.random.random((n_samples, problem.n_var))
        return (val < 0.5).astype(bool)


class IntegerRandomSampling(FloatRandomSampling):

    def _do(self, problem, n_samples, **kwargs):
        n, (xl, xu) = problem.n_var, problem.bounds()
        return np.column_stack([np.random.randint(xl[k], xu[k] + 1, size=n_samples) for k in range(n)])


class PermutationRandomSampling(Sampling):

    def _do(self, problem, n_samples, **kwargs):
        X = np.full((n_samples, problem.n_var), 0, dtype=int)
        for i in range(n_samples):
            X[i, :] = np.random.permutation(problem.n_var)
        return X
